import os
import subprocess

from collections import namedtuple
from queue import Queue
from threading import Thread
from typing import List, Sequence, Union

from xappt.config import log as logger


CommandResult = namedtuple("CommandResult", ["result", "stdout", "stderr"])


class PipeMonitor(Thread):
    def __init__(self, fd, queue):
        super().__init__()
        self._fd = fd
        self._queue = queue

    def run(self):
        for line in iter(self._fd.readline, ''):
            self._queue.put(line)

    def eof(self):
        return not self.is_alive() and self._queue.empty()


class CommandRunner(object):
    """ A wrapper around `subprocess.Popen` with basic environment manipulation.
    The results of a subprocess operation will be returned as a `CommandResult`
    object, which includes the return code, and stdout/stderr. Note that
    stdout/stderr will only be populated if the command was run silently.

    >>> c = CommandRunner(env={})
    >>> c.env_var_add("TEST", "1234")
    >>> c.env['TEST']
    '1234'
    >>> if os.name == "nt":
    ...     command = "set"
    ... elif os.name == "posix":
    ...     command = "printenv"
    >>> result = c.run(command, silent=True)
    >>> result.result == 0
    True
    >>> result.stdout.count("TEST=1234")
    1

    """
    def __init__(self, **kwargs):
        self.cwd = kwargs.get('cwd', os.getcwd())
        self.env = kwargs.get('env', os.environ.copy())

    def _split_path_var(self, key: str) -> List[str]:
        values = self.env.get(key, "").split(os.pathsep)
        return [v for v in values if len(v.strip())]

    def env_path_append(self, key: str, value: str):
        values = self._split_path_var(key)
        values.append(value)
        self.env[key] = os.pathsep.join(values)

    def env_path_prepend(self, key: str, value: str):
        values = self._split_path_var(key)
        values.insert(0, value)
        self.env[key] = os.pathsep.join(values)

    def env_var_add(self, key: str, value: str):
        self.env[key] = value

    def env_var_remove(self, key: str):
        try:
            del self.env[key]
        except KeyError:
            pass

    def run(self, command: Union[bytes, str, Sequence], **kwargs) -> CommandResult:
        env = kwargs.get('env', self.env)
        shell = kwargs.get('shell', False)

        subprocess_args = {
            'cwd': kwargs.get('cwd') or self.cwd,
            'env': env,
            'shell': shell,
            'universal_newlines': True,
            'encoding': 'utf8',
        }

        silent = kwargs.get('silent', True)
        if silent:
            subprocess_args['stdout'] = subprocess.PIPE
            subprocess_args['stderr'] = subprocess.PIPE

        logger.debug("Running command %s", str(command))
        logger.debug("Command environment %s", str(env))
        logger.debug("Command working directory %s", subprocess_args['cwd'])
        proc = subprocess.Popen(command, **subprocess_args)

        if silent:
            def io_fn_default(_: str):
                pass

            stdout_fn = kwargs.get('stdout_fn', io_fn_default)
            stderr_fn = kwargs.get('stderr_fn', io_fn_default)
            stdout = []
            stderr = []
            stdout_fn(" ".join(command))

            q_out = Queue()
            q_err = Queue()
            t_out = PipeMonitor(proc.stdout, q_out)
            t_err = PipeMonitor(proc.stderr, q_err)
            while proc.poll() is None:
                t_out.start()
                t_err.start()
                while not t_out.eof() or not t_err.eof():
                    while not q_out.empty():
                        line = q_out.get().rstrip()
                        stdout.append(line)
                        stdout_fn(line)
                    while not q_err.empty():
                        line = q_err.get().rstrip()
                        stderr.append(line)
                        stderr_fn(line)
                t_out.join()
                t_err.join()
            proc.stdout.close()
            proc.stderr.close()
            result = proc.returncode
            return CommandResult(result, "\n".join(stdout), "\n".join(stderr))
        else:
            proc.communicate()
            return CommandResult(proc.returncode, None, None)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



from __future__ import annotations
import abc

from typing import Optional, TYPE_CHECKING

from xappt.models.plugins.base import BasePlugin
if TYPE_CHECKING:
    from xappt.models.plugins.tool import BaseTool


class BaseInterface(BasePlugin, metaclass=abc.ABCMeta):
    @classmethod
    def collection(cls) -> str:
        return "interface"

    @abc.abstractmethod
    def invoke(self, plugin: BaseTool, **kwargs):
        pass

    @abc.abstractmethod
    def message(self, message: str):
        pass

    @abc.abstractmethod
    def warning(self, message: str):
        pass

    @abc.abstractmethod
    def error(self, message: str, *, details: Optional[str] = None):
        pass

    @abc.abstractmethod
    def ask(self, message: str) -> bool:
        pass

    @abc.abstractmethod
    def progress_start(self):
        pass

    @abc.abstractmethod
    def progress_update(self, message: str, percent_complete: float):
        pass

    @abc.abstractmethod
    def progress_end(self):
        pass



import xappt


@xappt.register_plugin(visible=False)
class ExamplePlugin(xappt.BaseTool):
    arg1 = xappt.ParamString(required=True)
    arg2 = xappt.ParamString(required=True)
    arg3 = xappt.ParamString(required=True)
    arg4 = xappt.ParamList(choices=['1', '2', '3', '4'])
    arg5 = xappt.ParamBool(default=False)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.arg5.on_value_changed.add(self._on_arg5_changed)

    def _on_arg5_changed(self, param: xappt.Parameter):
        if param.value:
            self.arg4.choices = ['5', '6', '7', '8']
        else:
            self.arg4.choices = ['1', '2', '3', '4']

    @classmethod
    def name(cls) -> str:
        return "example"

    @classmethod
    def help(cls) -> str:
        return "A simple command that will just echo the passed in arguments"

    def execute(self, **kwargs) -> int:
        self.interface.message(self.arg1.value)
        self.interface.message(self.arg2.value)
        self.interface.message(self.arg3.value)
        self.interface.message(self.arg4.value)
        self.interface.message(self.arg5.value)
        return 0



import os
import pathlib
import json

from collections import namedtuple
from typing import Any, Callable, List, Optional

ConfigItem = namedtuple("ConfigItem", ("key", "saver", "loader", "default"))


class ConfigMixin:
    def __init__(self):
        super().__init__()
        self.config_path: Optional[pathlib.Path] = None
        self._registry: List[ConfigItem] = []

    def add_config_item(self, key: str, *, saver: Callable, loader: Callable, default: Any = None):
        self._registry.append(ConfigItem(key, saver, loader, default))

    def _check_settings_file_name(self):
        if self.config_path is None:
            raise RuntimeError("`config_path` has not been set")

    def load_config(self):
        self._check_settings_file_name()

        loaded_settings_raw = {}
        try:
            with open(self.config_path, "r") as fp:
                try:
                    loaded_settings_raw = json.load(fp)
                except json.JSONDecodeError:
                    pass
        except FileNotFoundError:
            pass

        for item in self._registry:
            value = loaded_settings_raw.get(item.key, item.default)
            # noinspection PyBroadException
            try:
                item.loader(value)
            except BaseException:
                pass

    def save_config(self):
        self._check_settings_file_name()

        settings_dict = {}
        for item in self._registry:
            value = item.saver()
            settings_dict[item.key] = value

        os.makedirs(self.config_path.parent, exist_ok=True)
        with open(self.config_path, "w") as fp:
            json.dump(settings_dict, fp, indent=2)



from typing import Any

from xappt.models.mixins import ConfigMixin
from xappt.utilities.path import user_data_path


class BasePlugin(ConfigMixin):
    def __init__(self):
        super().__init__()
        self._data_dict = {}
        config_file = f"{self.collection()}-{self.name()}.cfg"
        self.config_path = user_data_path().joinpath("xappt").joinpath("plugins").joinpath(config_file)
        self.init_config()

    def init_config(self):
        self.add_config_item('data',
                             saver=lambda: self._data_dict,
                             loader=lambda x: self._data_dict.update(x),
                             default={})

    @classmethod
    def name(cls) -> str:
        """ Use this to specify the name of your plugin. """
        return cls.__name__.lower()

    @classmethod
    def help(cls) -> str:
        """ Provide a short help string or description. """
        return ""

    @classmethod
    def collection(cls) -> str:
        """ Use this to group your plugin with other like plugins. """
        return ""

    def data(self, key: str, default: Any = None) -> Any:
        self.load_config()
        return self._data_dict.get(key, default)

    def set_data(self, key: str, value: Any):
        self._data_dict[key] = value
        self.save_config()



import importlib
import inspect
import os
import sys

from functools import partial
from itertools import chain
from typing import Generator, Optional, Tuple, Type

from xappt.constants import *
from xappt.config import log as logger
from xappt.models import BaseTool, BaseInterface

__all__ = [
    'get_tool_plugin',
    'get_interface_plugin',
    'get_interface',
    'register_plugin',
    'discover_plugins',
    'registered_tools',
    'registered_interfaces'
]


PLUGIN_REGISTRY = {
    PLUGIN_TYPE_TOOL: {},
    PLUGIN_TYPE_INTERFACE: {},
}


def get_tool_plugin(plugin_name) -> Type[BaseTool]:
    plugin = PLUGIN_REGISTRY[PLUGIN_TYPE_TOOL].get(plugin_name)
    if plugin is None:
        raise ValueError("Tool Plugin '{}' not found".format(plugin_name))
    return plugin['class']


def get_interface_plugin(plugin_name) -> Type[BaseInterface]:
    plugin = PLUGIN_REGISTRY[PLUGIN_TYPE_INTERFACE].get(plugin_name)
    if plugin is None:
        raise ValueError("Interface Plugin '{}' not found".format(plugin_name))
    return plugin['class']


def get_interface(interface_name: Optional[str] = None) -> BaseInterface:
    if interface_name is None:
        interface_class = get_interface_plugin(os.environ.get(INTERFACE_ENV, INTERFACE_DEFAULT))
    else:
        interface_class = get_interface_plugin(interface_name)
    return interface_class()


def _add_plugin_to_registry(plugin_class, *, visible: bool):
    if issubclass(plugin_class, BaseTool):
        plugin_type = PLUGIN_TYPE_TOOL
    elif issubclass(plugin_class, BaseInterface):
        plugin_type = PLUGIN_TYPE_INTERFACE
    else:
        raise NotImplementedError

    plugin_name = plugin_class.name()

    if plugin_name in PLUGIN_REGISTRY[plugin_type]:
        logger.warning(f"a plugin with the name '{plugin_name}' has already been registered")
        return False

    if not len(plugin_class.collection().strip()):
        logger.warning(f"Invalid collection '{plugin_class.collection()}' for plugin '{plugin_name}'")
        return False

    logger.debug(f"registered plugin '{plugin_name}'")
    PLUGIN_REGISTRY[plugin_type][plugin_name] = {
        "class": plugin_class,
        "visible": visible,
    }

    return True


def find_plugin_modules(path: str) -> Generator[str, None, None]:
    for item in os.scandir(path):
        name_lower = item.name.lower()
        if not name_lower.startswith(PLUGIN_PREFIX):
            continue
        if name_lower.count(".dist-info") or name_lower.count(".egg-info"):
            continue
        yield item.name


def discover_plugins():
    logger.debug("discovering plugins")
    possible_plugin_modules = set()
    plugin_paths = set()

    env_paths = os.environ.get(PLUGIN_PATH_ENV, "").split(os.pathsep)
    if len(env_paths):
        logger.debug(f"{PLUGIN_PATH_ENV}: {os.pathsep.join(env_paths)}")

    for p in chain(env_paths, sys.path):
        if len(p) == 0 or not os.path.isdir(p):
            continue
        logger.debug(f"scanning path for plugins: {p}")
        for module in find_plugin_modules(p):
            if module in possible_plugin_modules:
                continue
            logger.debug(f"found possible plugin module: {module}")
            possible_plugin_modules.add(module)
            plugin_paths.add(p)

    for p in plugin_paths:
        if p not in sys.path:
            logger.debug(f"adding to sys.path: {p}")
            sys.path.append(p)

    for module in possible_plugin_modules:
        logger.debug(f"importing module: {module}")
        try:
            importlib.import_module(module)
        except ImportError:
            logger.debug(f"could not import {module}", exc_info=True)
        logger.debug(f"importing module {module}.plugins")
        try:
            importlib.import_module(f"{module}.plugins")
        except ImportError:
            logger.debug(f"could not import {module}.plugins", exc_info=True)


def register_plugin(cls=None, *, active=True, visible=True):
    if cls is None:
        return partial(register_plugin, active=active, visible=visible)
    assert inspect.isclass(cls)
    if active:
        _add_plugin_to_registry(cls, visible=visible)
    else:
        logger.debug(f"skipping inactive plugin: {cls.__name__}")
    return cls


def registered_tools(*, include_hidden=False) -> Generator[Tuple[str, Type[BaseTool]], None, None]:
    for tool_name, tool_dict in PLUGIN_REGISTRY[PLUGIN_TYPE_TOOL].items():
        if not tool_dict['visible'] and not include_hidden:
            continue
        yield tool_name, tool_dict['class']


def registered_interfaces(*, include_hidden=False) -> Generator[Tuple[str, Type[BaseInterface]], None, None]:
    for iface_name, iface_dict in PLUGIN_REGISTRY[PLUGIN_TYPE_INTERFACE].items():
        if not iface_dict['visible'] and not include_hidden:
            continue
        yield iface_name, iface_dict['class']
